# -*- coding: utf-8 -*-
"""
Device screenshot method.
"""
import os
import time
import traceback
from baseImage import Image, Rect
from base64 import b64decode
from .ui_driver import SIFT

import flybirds.core.global_resource as gr
import flybirds.utils.file_helper as file_helper
import flybirds.utils.flybirds_log as log
import flybirds.utils.uuid_helper as uuid_helper
from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.core.plugin.plugins.default.ios_snapshot import get_screen
from flybirds.core.exceptions import FlybirdsException


class BaseScreen:

    @staticmethod
    def screen_shot(path):
        """
        Take a screenshot and save
        """
        log.debug(f"[screen_shot] screen shot start. path is:{path}")
        cur_platform = g_Context.platform
        try:
            if cur_platform is None:
                log.error('[screen_shot] get cur_platform is None!')
                raise Exception("[screen_shot] get cur_platform is None!")

            poco = g_Context.ui_driver_instance
            screen_size = gr.get_device_size()
            if cur_platform.strip().lower() == "ios":
                b64img, fmt = get_screen()
            else:
                b64img, fmt = poco.snapshot(width=screen_size[1])

            open(path, "wb").write(b64decode(b64img))
        except Exception as e:
            log.warn(
                "Screenshot failed path: {}, error: {}".format(path, str(e)),
                traceback.format_exc(),
            )
        log.debug("[screen_shot] screen shot end!")

    @staticmethod
    def screen_link_to_behave(scenario, step_index, tag=None, link=True):
        """
        screenshot address and linked to the <scr> tag
        The label information is placed in the description of the scene,
        and the json report is processed after all the runs are finished,
        and the <scr> information in the description is converted into
        embeddings information in the step.
        """
        feature_name = file_helper.valid_file_name(scenario.feature.name)
        scenario_name = file_helper.valid_file_name(scenario.name)

        step_len = len(scenario.steps)
        if scenario is not None and hasattr(scenario, "background_steps") and \
                len(scenario.background_steps) > 0:
            step_len = step_len + len(scenario.background_steps)

        if step_len > step_index >= 0:
            file_name = None
            if not (tag is None):
                file_name = tag
            file_name += (
                    scenario_name
                    + uuid_helper.create_short_uuid()
                    + str(int(round(time.time() * 1000)))
                    + ".png"
            )

            screen_shot_dir = gr.get_screen_save_dir()
            if not (screen_shot_dir is None):
                current_screen_dir = os.path.join(screen_shot_dir,
                                                  feature_name)
            else:
                current_screen_dir = os.path.join(feature_name)
            log.debug(f"[screen_link_to_behave] screen_shot_dir path :"
                     f"{screen_shot_dir} and "
                     f"current_screen_dir path: {current_screen_dir}")
            file_helper.create_dirs_path_object(current_screen_dir)

            src_path = "../screenshot/{}/{}".format(feature_name, file_name)
            log.debug("[screen_link_to_behave] src_path: {}".format(src_path))
            data = (
                'embeddingsTags, stepIndex={}, <image class ="screenshot"'
                ' width="375" src="{}" />'.format(step_index, src_path)
            )
            if link is True:
                scenario.description.append(data)
            screen_path = os.path.join(current_screen_dir, file_name)
            g_Context.screen.screen_shot(screen_path)
            return screen_path

    @staticmethod
    def image_ocr(img_path):
        """
        Take a screenshot and ocr
        """
        log.debug(f"[image ocr path] image path is:{img_path}")
        ocr = g_Context.ocr_driver_instance

        if ocr is None:
            message = "\n----------------------------------------------------\n" \
                      "OCR engine is not start, please check following steps:\n"\
                      "1. In windows platform, you need to download OCR requirement file from " \
                      "https://github.com/ctripcorp/flybirds/blob/main/requirements_ml.txt\n" \
                      "2. run command `pip install -r requirements_ml.txt`\n" \
                      "3. Configure `ocrLang` option in flybirds_config.json, detail languages refer to \n" \
                      "https://flybirds.readthedocs.io/zh_CN/latest/BDD-UI-Testing-Flybirds.html#ocr-opencv\n" \
                      "----------------------------------------------------\n "
            raise FlybirdsException(message)

        g_Context.ocr_result = ocr.ocr(img_path, cls=True)
        g_Context.image_size = Image(img_path).size
        log.debug(f"[image ocr path] image size is:{g_Context.image_size}")


    @staticmethod
    def image_verify(img_source_path, img_search_path):
        """
        Take a screenshot and verify image
        """
        match = SIFT()
        img_source = Image(img_source_path)
        img_search = Image(img_search_path)

        result = match.find_all_results(img_source, img_search)
        return result

    @staticmethod
    def white_screen_detect(img_path):
        match = SIFT()
        img = Image(img_path)
        start_time = time.time()
        width = img.size[1]
        height = img.size[0]
        range_height = height / 100
        point_y = 0
        white_percent = 0
        while point_y + range_height <= height:
            new_img = img.crop(rect=Rect(0, point_y, width, range_height))
            kp_src, des_src = match.get_keypoint_and_descriptor(new_img)
            point_y += range_height
            if len(kp_src) < 3:
                white_percent += 1
        log.info(f"detect use time:{time.time() - start_time}")
        return white_percent






# -*- coding: utf-8 -*-
import operator
import os
import traceback
from subprocess import Popen

from flybirds.core.config_manage import DeviceConfig
from flybirds.report.fail_feature_create import rerun_launch
from flybirds.report.parallel_runner import parallel_run
from flybirds.utils import flybirds_log as log
from flybirds.utils.dsl_helper import get_use_define_param
from flybirds.utils.pkg_helper import load_pkg_by_ns


class RunManage:
    """
    run cycle manage class
    """

    def __init__(self):
        pass

    before_run_processor = []
    after_run_processor = []

    @staticmethod
    def load_pkg():
        """
        load report ref
        """
        load_pkg_by_ns("flybirds.core.launch_cycle")
        load_pkg_by_ns("flybirds.report.gen")

    @staticmethod
    def process(processors_name, *args):
        """
        processor executor
        """
        processors = getattr(RunManage, processors_name)
        if processors is not None and len(processors) > 0:
            sort_key = operator.attrgetter("order")
            processors.sort(key=sort_key)
            for processor in processors:
                if hasattr(processor, "can"):
                    can_run = processor.can(*args)
                    if can_run is True:
                        processor.run(*args)
                else:
                    processor.run(*args)

    @staticmethod
    def run(context):
        """
        run logical contains : before ,exe ,after
        """
        RunManage.load_pkg()
        RunManage.process("before_run_processor", context)
        RunManage.exe(context)
        RunManage.process("after_run_processor", context)

    @staticmethod
    def exe(context):
        """
        start behave process and rerun fail case
        """
        no_args = context.get("no_args")
        if no_args is not None and no_args is False:

            # noinspection PyBroadException
            try:
                # parallel.run(context)
                is_parallel = need_parallel_run(context)
                if is_parallel:
                    parallel_run(context)
                else:
                    cmd_str = context.get("cmd_str")
                    behave_process = Popen(
                        cmd_str, cwd=os.getcwd(), shell=True, stdout=None
                    )
                    behave_process.wait()
                    behave_process.communicate()

                rerun_launch(context, is_parallel)

            except Exception:
                log.error(
                    f"behave task execute error: {traceback.format_exc()}")
        else:
            log.info("cannot run non-args")

    @staticmethod
    def join(processors_name, processor, replace=0):
        """
        add processor
        """
        if processor is not None and processor.name is not None:
            processors = getattr(RunManage, processors_name)
            indx = -1
            for i, item in enumerate(processors):
                if item.name is not None and item.name == processor.name:
                    indx = i
                    break
            if replace == 0:
                if indx >= 0:
                    return indx
            elif indx >= 0:
                processors[indx] = processor
                return indx
            processors.append(processor)

            return indx
        return -1

    @staticmethod
    def insert(processors_name, processor, replace=0):
        """
        insert process
        """
        if processor is not None and processor.name is not None:
            processors = getattr(RunManage, processors_name)
            indx = -1
            for i, item in enumerate(processors):
                if item.name is not None and item.name == processor.name:
                    indx = i
                    break
            if replace == 0:
                if indx >= 0:
                    return indx
            elif indx >= 0:
                del processors[indx]
            processors.insert(0, processor)

            return indx
        return -1


def run_script(run_args):
    log.info(f"received run_args: {run_args}")
    try:
        r_context = {
            "run_args": run_args
        }
        RunManage.run(r_context)
    except Exception:
        log.error(f"behave task run error: {traceback.format_exc()}")


def need_parallel_run(context):
    user_data = get_use_define_param(context, 'platform')
    platform = DeviceConfig(user_data, None).platform
    # check platform
    if platform not in ['ios', 'android', 'web']:
        log.warn(f'flybirds is not supports to run on {platform} '
                 f'platform. It will now run on Android by default.')
        platform = "android"
    context['cur_platform'] = platform
    if 'web' == platform.lower():
        return True
    return False



# -*- coding: utf-8 -*-
"""
web screen impl
"""

import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log
from flybirds.core.plugin.plugins.default.screen import BaseScreen

__open__ = ["Screen"]


class Screen(BaseScreen):
    """
    web screen impl
    """
    name = "web_screen"

    @staticmethod
    def screen_shot(path):
        log.info(f"[web screen_shot] screen shot start. path is:{path}")
        page_obj = gr.get_value("plugin_page")
        if page_obj is None or (not hasattr(page_obj, 'page')):
            log.error('[web_screenshot] get page object has error!')
        page_obj.page.screenshot(path=f'{path}')
        log.info("[web screen_shot] screen shot end!")



# -*- coding: utf-8 -*-
"""
Poco manage api
"""
import flybirds.core.global_resource as gr


def ocr_init(lang=None):
    """
    Initialize the paddleocr object
     :return:
    """
    if lang is None:
        ocr_lang = gr.get_app_config_value("ocr_lang")
    else:
        ocr_lang = lang

    if ocr_lang != "":
        from paddleocr import PaddleOCR
        # Paddleocr support languages
        # example`ch`, `en`, `fr`, `german`, `korean`, `japan`
        det_limit_type = gr.get_frame_config_value("ocr_det_limit_type")
        det_limit_side_len = gr.get_frame_config_value("ocr_det_limit_side_len")
        ocr = PaddleOCR(use_angle_cls=True,
                        lang=ocr_lang,
                        det_limit_type=det_limit_type,
                        det_limit_side_len=det_limit_side_len)
        # need to run only once to download and load model into memory

        return ocr




# -*- coding: utf-8 -*-
"""
Swipe apis
"""
import time

import flybirds.core.global_resource as gr
import flybirds.core.plugin.plugins.default.ui_driver.poco.findsnap as findsnap
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_ele as poco_ele
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_manage as pm
import flybirds.utils.point_helper as point_helper
from flybirds.core.exceptions import FlybirdNotFoundException
from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils import language_helper as lan
from flybirds.core.plugin.plugins.default.step.verify import ocr, ocr_txt_contain, img_exist
import flybirds.utils.flybirds_log as log


def air_bdd_full_screen_swipe(
    poco,
    start_point,
    screen_size,
    direction,
    distance,
    duration,
    ready_time=None,
):
    """
    In the full screen range, slide from the specified starting point in one
    direction up, down, left, and right to specify the distance.
    """
    # get current language
    language = g_Context.get_current_language()
    direct_left = lan.parse_glb_str("left", language)
    direct_right = lan.parse_glb_str("right", language)

    if not (ready_time is None):
        time.sleep(ready_time)
    if start_point[0] > 1:
        start_point[0] = start_point[0] / screen_size[0]
    if start_point[1] > 1:
        start_point[1] = start_point[1] / screen_size[1]
    if distance > 1:
        if direction == direct_left or direction == direct_right:
            distance /= screen_size[0]
        else:
            distance /= screen_size[1]
    air_bdd_direction_swipe(poco, start_point, direction, distance, duration)


def air_bdd_ele_swipe(
    poco,
    container_dsl_str,
    optional,
    start_point,
    screen_size,
    direction,
    distance,
    duration,
    ready_time=None,
):
    """
    Slides a specified distance up, down, left, or right from a specified
    starting point within a sliding element.
    """
    poco_ele.wait_exists(poco, container_dsl_str, optional)

    if not (ready_time is None):
        time.sleep(ready_time)

    poco_object = pm.create_poco_object_by_dsl(
        poco, container_dsl_str, optional
    )
    target_position = poco_object.get_position()
    target_size = poco_object.get_size()
    if start_point[0] > 1:
        start_point[0] = (
            (target_position[0] * screen_size[0])
            - (target_size[0] / 2 * screen_size[0])
            + start_point[0]
        ) / screen_size[0]
    else:
        start_point[0] = (
            target_position[0]
            - target_size[0] / 2
            + target_size[0] * start_point[0]
        )
    if start_point[1] > 1:
        start_point[1] = (
            (target_position[1] * screen_size[1])
            - (target_size[1] / 2 * screen_size[1])
            + start_point[1]
        ) / screen_size[1]
    else:
        start_point[1] = (
            target_position[1]
            - target_size[1] / 2
            + target_size[1] * start_point[1]
        )

    # get current language
    language = g_Context.get_current_language()
    direct_left = lan.parse_glb_str("left", language)
    direct_right = lan.parse_glb_str("right", language)

    if distance > 1:
        if direction == direct_left or direction == direct_right:
            distance /= screen_size[0]
        else:
            distance /= screen_size[1]
    else:
        if direction == direct_left or direction == direct_right:
            distance *= target_size[0]
        else:
            distance *= target_size[1]

    air_bdd_direction_swipe(poco, start_point, direction, distance, duration)


def air_bdd_direction_swipe(
    poco, start_point, direction, distance, duration=None
):
    """
    swipe the specified distance from the starting point to one of up, down,
    left, and right.
    :param poco:
    :param start_point:
    :param direction:
    :param distance:
    :param duration:
    :return:
    """
    # get current language
    language = g_Context.get_current_language()
    direct_left = lan.parse_glb_str("left", language)
    direct_right = lan.parse_glb_str("right", language)
    direct_up = lan.parse_glb_str("up", language)
    direct_down = lan.parse_glb_str("down", language)

    end_point = [start_point[0], start_point[1]]
    if direction == direct_left:
        end_point[0] -= distance
        if end_point[0] < 0.0:
            end_point[0] = 0
            start_point[0] = distance
    elif direction == direct_right:
        end_point[0] += distance
        if end_point[0] > 1.0:
            end_point[0] = 1
            start_point[0] = 1 - distance
    elif direction == direct_up:
        end_point[1] -= distance
        if end_point[1] < 0:
            end_point[1] = 0
            start_point[1] = distance
    elif direction == direct_down:
        end_point[1] += distance
        if end_point[1] > 1:
            end_point[1] = 1
            start_point[1] = 1 - distance
    air_bdd_percent_point_swipe(poco, start_point, end_point, duration)


def air_bdd_percent_point_swipe(poco, start_point, end_point, duration=None):
    """
    swipe from the start point to the end point
    :param poco:
    :param start_point:
    :param end_point:
    :param duration:
    :return:
    """
    if duration is None:
        poco.swipe(start_point, end_point)
    else:
        poco.swipe(start_point, end_point, duration=duration)
    if gr.get_frame_config_value("use_snap", False):
        findsnap.fix_refresh_status(True)


def full_screen_swipe_search(
    poco,
    search_dsl_str,
    search_optional,
    swipe_count,
    direction,
    screen_size,
    start_x=None,
    start_y=None,
    distance=None,
    duration=None,
):
    """
    Full screen swipe to find
    """
    direction = point_helper.search_direction_switch(direction)
    start_point = point_helper.get_swipe_search_start_point(
        direction, start_x, start_y
    )

    if distance is None:
        distance = 0.3

    log_count = swipe_count
    searched = False
    while swipe_count >= 0:
        print("swipe_count", swipe_count)
        try:
            search_poco_object = pm.create_poco_object_by_dsl(
                poco, search_dsl_str, search_optional
            )
            if search_poco_object.exists():
                searched = True
                break
        except Exception:
            pass
        if swipe_count == 0:
            break
        air_bdd_full_screen_swipe(
            poco, start_point, screen_size, direction, distance, duration
        )
        swipe_count -= 1
    if not searched:
        message = "swipe to {} {} times，not find {}".format(
            direction, log_count, search_dsl_str
        )
        raise FlybirdNotFoundException(message, {})
    if gr.get_frame_config_value("use_snap", False):
        findsnap.fix_refresh_status(True)


def air_bdd_swipe_search(
    poco,
    container_dsl_str,
    container_optional,
    search_dsl_str,
    search_optional,
    swipe_count,
    screen_size,
    direction,
    start_x=None,
    start_y=None,
    distance=None,
    duration=None,
):
    """
    swipe in a certain direction in the full screen or within an element to
    find an element.
    """
    poco_ele.wait_exists(poco, container_dsl_str, container_optional)
    direction = point_helper.search_direction_switch(direction)
    start_point = point_helper.get_swipe_search_start_point(
        direction, start_x, start_y
    )

    if distance is None:
        distance = 0.3

    log_count = swipe_count
    searched = False
    while swipe_count >= 0:
        try:
            search_poco_object = pm.create_poco_object_by_dsl(
                poco, search_dsl_str, search_optional
            )
            if search_poco_object.exists():
                searched = True
                break
        except Exception:
            pass
        if swipe_count == 0:
            break
        air_bdd_ele_swipe(
            poco,
            container_dsl_str,
            container_optional,
            start_point,
            screen_size,
            direction,
            distance,
            duration,
        )
        swipe_count -= 1
    if not searched:
        message = "{} swipe to {} {} times，not find {}".format(
            container_dsl_str, direction, log_count, search_dsl_str
        )
        raise FlybirdNotFoundException(message, {})
    if gr.get_frame_config_value("use_snap", False):
        findsnap.fix_refresh_status(True)


def full_screen_swipe_search_ocr(
    context,
    poco,
    search_dsl_str,
    swipe_count,
    direction,
    screen_size,
    start_x=None,
    start_y=None,
    distance=None,
    duration=None,
):
    """
    Full screen swipe to find
    """
    direction = point_helper.search_direction_switch(direction)
    start_point = point_helper.get_swipe_search_start_point(
        direction, start_x, start_y
    )

    if distance is None:
        distance = 0.3

    log_count = swipe_count
    searched = False
    while swipe_count >= 0:
        try:
            ocr(context)
            searched = ocr_txt_contain(context, search_dsl_str, islog=False)
            if searched is True:
                break
        except Exception:
            pass
        if swipe_count == 0:
            break
        air_bdd_full_screen_swipe(
            poco, start_point, screen_size, direction, distance, duration
        )
        swipe_count -= 1
    if not searched:
        for line in g_Context.ocr_result:
            log.info(f"[image ocr result] scan line info is:{line}")
        message = "swipe to {} {} times，not find {}".format(
            direction, log_count, search_dsl_str
        )
        raise FlybirdNotFoundException(message, {})


def full_screen_swipe_search_img(
    context,
    poco,
    search_dsl_str,
    swipe_count,
    direction,
    screen_size,
    start_x=None,
    start_y=None,
    distance=None,
    duration=None,
):
    """
    Full screen swipe to find
    """
    direction = point_helper.search_direction_switch(direction)
    start_point = point_helper.get_swipe_search_start_point(
        direction, start_x, start_y
    )

    if distance is None:
        distance = 0.3

    log_count = swipe_count
    searched = False
    while swipe_count >= 0:
        try:
            searched = img_exist(context, search_dsl_str, islog=False)
            if searched is True:
                break
        except Exception:
            pass
        if swipe_count == 0:
            break
        air_bdd_full_screen_swipe(
            poco, start_point, screen_size, direction, distance, duration
        )
        swipe_count -= 1
    if not searched:
        step_index = context.cur_step_index - 1
        src_path = "../../../{}".format(search_dsl_str)
        data = (
            'embeddingsTags, stepIndex={}, <image class ="screenshot"'
            ' width="375" src="{}" />'.format(step_index, src_path)
        )
        context.scenario.description.append(data)

        message = "swipe to {} {} times，not find {}".format(
            direction, log_count, search_dsl_str
        )
        raise FlybirdNotFoundException(message, {})




# -*- coding: utf-8 -*-
"""
This module contains the device Core APIs.
"""
import flybirds.utils.flybirds_log as log

from flybirds.core.global_context import GlobalContext as g_Context


def device_connect(device_id):
    """
    Initialize device with uri, and set as current device.
    :param device_id: device id
    :return: device instance
    """
    dev = g_Context.device.device_connect(device_id)
    log.info("device connect info:{}".format(dev))
    return dev


def use_shell(cmd):
    """
    Start remote shell in the target device and execute the command
      :platforms: Android
    :param cmd:  command to be run on device, e.g. "ls /data/local/tmp"
    :return:
    """
    return g_Context.device.use_shell(cmd)
